/* menufunc.c: menu functions
 *
 * Copyright (C) 1995-97 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
 *               1996-97 Michael Schlueter <schlue00@marvin.informatik.uni-dortmund.de>
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation: either version 2 or
 * (at your option) any later version.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#include "fdisk.h"
#include "partable.h"
#include "writepart.h"
#include "input.h"
#include "util.h"


void set_bootable( void )
{
    int i, j, n_ext, dflt, max = n_bootflags;
    int *pflag;

    if (!partitions) {
	printf( "No partitions!\n" );
	return;
    }
    table_infos( NULL, &n_ext, NULL, NULL );
    if (n_ext > 0) {
	i = read_int( 0, partitions, partitions, ignore,
		      "Partition number (0 = XGM)" );
    }
    else {
	i = (partitions == 1) ? 1 : read_int( 1, partitions, partitions,
					      ignore, "Partition number" );
    }
    --i;
    pflag = (i < 0) ? &XGM_flag : &(part_table[i].flag);
    
    printf( "\nCurrent boot flag is: " );
    if (*pflag) {
	printf( "0x%02x %s\n", *pflag, atari_boot_type( *pflag) );
    }
    else {
	printf( "not bootable\n" );
    }
    
    if (i >= 0 && IS_EXTENDED(part_table[i]))
	printf( "\nWarning: Bootflags usually not supported on extended "
		"partitions!\n" );
    
    dflt = 0;
    printf( "\nPossible boot flags are:\n" );
    printf( "  0: not bootable\n" );
    for( j = 0; j < max; ++j ) {
	printf( "  %d: 0x%02x %s\n",
		j+1, bootflags[j].flag, bootflags[j].name );
	if (bootflags[j].flag == *pflag)
	    dflt = j+1;
    }
    j = read_int( 0, dflt, max, deflt, "New flag");
    *pflag = j > 0 ? bootflags[j-1].flag : 0;
}

void bad_sector_list( void )
{
    
    if (bsl_size == 0) {
	printf( "There is no bad sector list currently (start=%lu size=0)\n",
		bsl_start );
    }
    else {
	printf( "The bad sector list resides from %lu to %lu (size %lu)\n",
		bsl_start, bsl_start+bsl_size-1, bsl_size );
    }

    printf( "  1: remove bad sector list\n" );
    printf( "  2: define bad sector list\n" );
    printf( "  3: create HDX compatible empty bad sector list\n" );
    printf( "     (Note: Only necessary if you want to use also HDX on this disk)\n" );
    
    switch( read_int( 1, 3, 3, ignore, "Action") ) {

      case 1:
	bsl_start = bsl_size = 0;
	bsl_HDX_compat = 0;
	break;
	
      case 2:
      { unsigned long old_start, old_size;
	unsigned long start, def_start, end, max_size;

	/* free before new allocation */
	old_start = bsl_start;
	old_size  = bsl_size;
	bsl_start = bsl_size = 0;

	def_start = 1 + allocated_sectors_at( 1 );
	if (def_start >= hd_size) {
	    printf( "There are no unallocated sectors!\n" );
	    bsl_start = old_start;
	    bsl_size  = old_size;
	    return;
	}
	for(;;) {
	    start = read_int( 1, def_start, hd_size-1, deflt,
			      "First sector" );
	    if ((max_size = free_sectors_at( start )) > 0)
		break;
	    printf( "Sector %lu is not free!\n", start );
	}
	end = read_int( start, start+1, start+max_size-1, lower,
			"Last sector or +size or +sizeM or +sizeK" );
	bsl_size = end+1 - start;
	bsl_start = start;
	bsl_HDX_compat = 0;
	break;
      }
	
      case 3:
      { unsigned long old_start, old_size;

        old_start = bsl_start;
	old_size  = bsl_size;
	bsl_start = bsl_size = 0;

	/* use first unallocated sector */
	bsl_start = 1 + allocated_sectors_at( 1 );
	if (bsl_start >= hd_size) {
	    printf( "There are no unallocated sectors!\n" );
	    bsl_start = old_start;
	    bsl_size  = old_size;
	    return;
	}
	bsl_size = 1;
	bsl_HDX_compat = 1;
	break;
      }
    }
}

void change_partition_size( int i )
{
    long start, end, roots, free_before, free_after, new_size;
    PARTITION *p = &part_table[i];

    if (!expert_mode && p->contents_valid) {
	if (!read_yesno( "This will make the partition contents invalid. "
			 "Continue", 0 ))
	    return;
    }
    
    start = p->start;
    end   = start + p->size;
    roots = p->rootsec;

    free_before = free_sectors_before( start );
    /* if free area is limited by this partition's ars, search again */
    if (start - free_before - 1 == roots)
	free_before += 1 + free_sectors_before( roots );

    free_after = free_sectors_at( end );

    start = read_int( start - free_before,	/* lower   */
		      start,			/* default */
		      end-1,			/* max     */
		      deflt,			/* relto   */
		      "New first sector or [+-]size[MK]" );

    end = read_int( start,			/* lower   */
		    end - 1,			/* default */
		    end + free_after - 1,	/* max     */
		    deflt,				/* relto   */
		    "New last sector or [+-]size[MK]" ) + 1;

    new_size = end - start;
    p->start = start;
    p->size  = new_size;
    if (roots >= p->start) {
	p->rootsec = p->start;
	p->start++;
	new_size--;
	if (expert_mode)
	    printf( "Auxiliary root sector now at %lu\n", p->rootsec );
    }
    
    printf( "New size: %lu sectors = %s MB\n",
	    new_size, round_mb( new_size ) );
    p->contents_valid = 0;
    if (new_size < 1024)
	printf( "Isn't that partition now rather small?\n" );
}

void delete_partition( int i )
{
    memmove( &part_table[i], &part_table[i+1],
	     (partitions-1-i)*sizeof(PARTITION) );
    partitions--;
}

void convert_partition( int i )
{
    int n_prim, n_ext, first_ext, last_ext;
    PARTITION *p = &part_table[i];
    int is_ext = IS_EXTENDED(*p);
    
    /* can convert only partitions that are at the "edge" of the XGM area;
     * furthermore, the very first partition must be primary, as the XGM may
     * not reside in the first slot */

    table_infos( &n_prim, &n_ext, &first_ext, &last_ext );

    if (is_ext) {
	/* convert extended -> primary */
	if (n_prim + (n_ext > 1) >= 4) {
	    printf( "No free slots for more primary partitions\n" );
	    return;
	}
	if (!(i == first_ext || i == last_ext)) {
	    printf( "Can only convert partitions at the edge of the "
		    "extended area\n" );
	    return;
	}

	if (p->rootsec == p->start - 1 && !p->contents_valid) {
	    p->start--;
	    p->size++;
	}
	else
	    printf( "Cannot integrate aux. root sector into partition. "
		    "It's unassigned now.\n" );
	p->rootsec = 0;
    }
    else {
	/* convert primary -> extended */
	if (i == 0) {
	    printf( "First partition must always be primary, "
		    "cannot convert\n" );
	    return;
	}
	if (n_ext > 0 && !(i == first_ext-1 || i == last_ext+1)) {
	    printf( "Can only convert partitions next to the existing "
		    "extended area\n" );
	    return;
	}
	if (!expert_mode && p->contents_valid) {
	    if (!read_yesno( "The conversion will make the partition "
			     "contents invalid.\nContinue", 0 ))
		return;
	}

	p->rootsec = p->start;
	p->start++;
	p->size--;
    }
}

void select_xfmt( void )
{
    printf( "Sorry, this function is currently still a dummy.\n"
	    "Currently, only AHDI format is supported.\n" );
}

void change_contents_flag( int i )
{
    PARTITION *p = &part_table[i];

    if (!p->contents_valid)
	p->contents_valid = 1;
    else {
	printf( "Partition %d's contents are currently marked valid\n", i+1 );
	if (!read_yesno( "Do you really want to mark them invalid", -1 ))
	    return;
	p->contents_valid = 0;
    }
    
    printf( "Partition %d's contents are now marked %svalid\n",
	    i+1, p->contents_valid ? "" : "in" );
}

void new_partition( void )
{
    PARTITION *p;
    unsigned long sec, n, start, def_start, min_start, max_start;
    unsigned long end, size, max_size;
    int header_printed, i, new_num;
    int n_prim, n_ext, first_ext, last_ext;
    int make_ext, prim_possible, ext_possible;
    char *why;
    
    if (partitions >= MAXIMUM_PARTS) {
	printf( "The maximum number of partitions has been created\n" );
	return;
    }

    /* Scan partition table for some data */
    table_infos( &n_prim, &n_ext, &first_ext, &last_ext );

    /* print list of free areas */
    header_printed = 0;
    def_start = min_start = max_start = 0;
    n = 0;
    for( sec = 1; sec < hd_size; sec += n ) {
	sec += allocated_sectors_at( sec );
	if (sec >= hd_size || !(n = free_sectors_at( sec )))
	    break;
	if (!header_printed) {
	    printf( "Free areas on disk are:\n" );
	    printf( "    Begin       End    Blocks     MB\n" );
	    header_printed = 1;
	}
	printf( "%9lu %9lu %9lu%c  %s\n",
		sec, sec+n-1, n/2, (n&1) ? '+' : ' ', round_mb(n) );
	if (!def_start) {
	    new_num = canonical_number( sec );
	    prim_possible = primary_possible( new_num, n_prim, n_ext,
					      first_ext, last_ext, NULL );
	    ext_possible = extended_possible( new_num, n_prim, n_ext,
					      first_ext, last_ext, NULL );
	    if (prim_possible || ext_possible)
		def_start = sec;
	}
	if (!min_start)
	    min_start = sec;
	max_start = sec+n-1;
    }
    if (!min_start) {
	printf( "There is no unallocated space on the disk!\n" );
	return;
    }
    if (!def_start)
	def_start = min_start;

    /* Ask user for start of new partition */
    for(;;) {
	start = read_int( min_start, def_start, max_start, deflt,
			  "First sector" );
	if ((max_size = free_sectors_at( start )) > 0)
	    break;
	printf( "Sector %lu is not free!\n", start );
    }
    /* ...and for the end */
    end = read_int( start, start+max_size-1, start+max_size-1, lower,
		    "Last sector or +size or +sizeM or +sizeK" );
    size = end+1 - start;

    if (!expert_mode) {
	/* In non-expert mode, new partitions always have canonical numbering
	 * (same order as data on disk). If it's not possible to create the
	 * partition there (no primary slot free and not in or next to
	 * XGM area), the create simply fails. */
	new_num = canonical_number( start );
	printf( "New partition will be number %d\n", new_num+1 );

	prim_possible = primary_possible( new_num, n_prim, n_ext,
					  first_ext, last_ext, NULL );
	ext_possible = extended_possible( new_num, n_prim, n_ext,
					  first_ext, last_ext, NULL );

	if (!prim_possible && !ext_possible) {
	    printf( "No more slots available in primary partition table, "
		    "or this partition \n"
		    "can't be an extended one, or would break XGM chain. "
		    "Try converting some\n"
		    "primary partition to extended, deleting some other "
		    "partitions or use\n"
		    "expert mode to solve this.\n" ); 
	    /* TODO: Maybe automatically convert prim->ext for a partition
	     * where the "contents valid" flag is cleared. */
	    return;
	}

	if (partitions > 0)
	    memmove( &part_table[new_num+1], &part_table[new_num],
		     (partitions-new_num)*sizeof(PARTITION) );
	++partitions;
	p = &part_table[new_num];

	p->start = start;
	p->size  = size;
	p->rootsec = 0;

	/* Create an extended partition if this is the 4th part. (to avoid
	 * prim->ext conversion later), or if already adjacent to XGM area. */
	if ((n_prim == 3 && n_ext == 0) ||
	    (n_ext > 0 && first_ext-1 <= new_num && new_num <= last_ext+1)) {
	    p->rootsec = start;
	    p->start++;
	    p->size--;
	}
    }
    else {
	/* In expert mode, we allow the user to give the partition an
	 * arbitrary number, choose its type (prim/ext), and to locate the aux
	 * root sector freely if extended. */

	for(;;) {

	    printf( "The following numbers are possible for this "
		    "partition:\n" );
	    for( i = 0; i <= partitions; ++i ) {
		prim_possible = primary_possible( i, n_prim, n_ext, first_ext,
						  last_ext, NULL );
		ext_possible = extended_possible( i, n_prim, n_ext, first_ext,
						  last_ext, NULL );
		if (prim_possible || ext_possible)
		    printf( "  %d(%s%s%s)\n", i+1,
			    prim_possible ? "p" : "",
			    prim_possible && ext_possible ? "," : "",
			    ext_possible ? "e" : "" );
	    }
	    
	    new_num = read_int( 0, partitions+1, partitions+1, ignore,
				"Partition number (0 = abort)" ) - 1;
	    if (new_num < 0)
		return;

	    prim_possible = primary_possible( new_num, n_prim, n_ext,
					      first_ext, last_ext, &why );
	    if (!prim_possible)
		printf( "Cannot be primary (%s)\n", why );

	    ext_possible = extended_possible( new_num, n_prim, n_ext,
					      first_ext, last_ext, &why );
	    if (!ext_possible)
		printf( "Cannot be extended (%s)\n", why );
	    
	    if (prim_possible || ext_possible)
		break;
	    printf( "This number is not possible for this partition\n" );
	}

	if (prim_possible && ext_possible) {
	    make_ext = read_yesno( "Make an extended partition?", 0 );
	}
	else if (prim_possible) {
	    printf( "This will be a primary partition\n" );
	    make_ext = 0;
	}
	else if (ext_possible) {
	    printf( "This will be an extended partition\n" );
	    make_ext = 1;
	}

	if (make_ext) {
	    for(;;) {
		n = read_int( 0, start, start, upper,
			      "Location for auxiliary root sector (0=abort)" );
		if (n == 0)
		    return;
		if (n == start || free_sectors_at( n ) > 0)
		    break;
		printf( "Sector %ld allocated and not at beginning of new "
			"partition\n", n );
	    }
	}

	if (partitions > 0)
	    memmove( &part_table[new_num+1], &part_table[new_num],
		     (partitions-new_num)*sizeof(PARTITION) );
	++partitions;
	p = &part_table[new_num];

	p->start = start;
	p->size  = size;
	p->rootsec = 0;

	if (make_ext) {
	    if (n == start) {
		p->start++;
		p->size--;
	    }
	    p->rootsec = n;
	}
    }

    /* common settings for new partition */
    strcpy( p->id, "LNX" );
    p->flag = 0;
    p->contents_valid = 0;
}

void move_partition(void)
{
    printf( "Not yet implemented, sorry.\n" );
}

void list_table(void)
{
    PARTITION *p;
    PARTITION XGM_part;
    char *type, numbuf[4];
    int i, slotno = 0, xgm_printed = -1;
    int w = strlen(disk_device);

    if (expert_mode || XGM_flag) {
	XGM_boundary( &XGM_part.start, &XGM_part.size );
	XGM_part.flag = XGM_flag;
	XGM_part.rootsec = 0;
	XGM_part.contents_valid = 0;
	strcpy( XGM_part.id, "XGM" );
	xgm_printed = 0;
    }
    
    printf( "\nDisk %s: %ld * %d bytes, %ld%s kBytes, %s MBytes\n",
	    disk_device, hd_size, SECTOR_SIZE,
	    hd_size/2, (hd_size&1) ? "+" : "",
	    round_mb( hd_size ) );
    if (w < 5)
	w = 5;
    printf("%*s  Boot Val Rootsec.    Begin      End   Blocks  Id   System\n",
	   w + 1, "Device");
    for( i = 0 ; i < partitions; i++ ) {
	p = &part_table[i];
	if (IS_EXTENDED(*p) && xgm_printed == 0) {
	    p = &XGM_part;
	    xgm_printed = 1;
	    --i;
	}

	if (p == &XGM_part)
	    strcpy( numbuf, "x " );
	else
	    sprintf( numbuf, "%-2d", i+1 );
	
	printf( expert_mode && p->rootsec == 0 ?
	    "%*s%s %3s   %c (slot %d)0%9ld%9ld%9ld%c %3s  %s\n" :
	    "%*s%s %3s   %c %9ld%9ld%9ld%9ld%c %3s  %s\n",
/* device          */	w, disk_device, numbuf,
/* boot flag       */	p->flag ? atari_boot_type_short(p->flag) : "",
/* contents valid  */	p == &XGM_part ? '-' : p->contents_valid ? '*' : '-',
/* rootsec         */	!expert_mode || p->rootsec ? p->rootsec : slotno++,
/* begin           */	p->start,
/* end             */	p->start + p->size - 1,
/* blocks          */	p->size / 2,
/* odd flag on end */	(p->size & 1) ? '+' : ' ',
/* type id         */	p->id,
/* type name       */	(type = atari_partition_type(p->id)) ? type :
			    p == &XGM_part ? "" : "Unknown"
	    ); 
    }
}

void change_partid( int i )
{
    int cnr;
    PARTITION *p = &part_table[i];

    for(;;) {
	list_atari_types();
	printf( "ID is %s. Enter new ID (3 letters): ", p->id );
	read_line();
	cnr = strlen( line_ptr );
	if (1 <= cnr && cnr <= 4 && isalnum(line_ptr[0]) &&
	    isalnum(line_ptr[1]) && isalnum(line_ptr[2]) )
	    break;
	printf( "ID must be three letters or digits\n" );
    }
    
    memcpy( &p->id, line_ptr, 3 );
    printf( "ID changed to %s\n", p->id );

    if (!(isupper(line_ptr[0]) &&
	  isupper(line_ptr[1]) &&
	  isupper(line_ptr[2])))
	printf( "(Don't you want to use three uppercase letters, "
		"which is the standard?)\n" );
}

#if 0
void select_options( void )
{
    int i;
    int n = n_global_options;
    
    printf( "Current option settings:\n" );
    for( i = 0; i < n; ++i ) {
	printf( "  %d) %s: %s\n", i+1,
		global_options[i].name,
		*global_options[i].flag ? "on" : "off" );
	printf( "    %s\n", global_options[i].description );
    }

    while( (i = read_int( 0, 0, n, ignore, "Change option (0 = quit)" )) ) {
	*global_options[i-1].flag ^= 1;
	printf( "%s is now %s\n",
		global_options[i].name,
		*global_options[i].flag ? "on" : "off" );
    }
}

void parse_options( char *options )
{
    char *p;
    int i, value;
    
    for( p = strtok( options, "," ); p; p = strtok( NULL, "," ) ) {
	if (strncmp( p, "no-", 3 ) == 0) {
	    value = 0;
	    p += 3;
	}
	else
	    value = 1;

	for( i = 0; i < n_global_options; ++i ) {
	    if (strcmp( global_options[i].name, p ) == 0)
		break;
	}
	if (i == n_global_options) {
	    fprintf( stderr, "Unrecognized option %s\n", p );
	    continue;
	}

	*global_options[i].flag = value;
    }
}
#endif

/* Do some consistency checks */
void verify(void)
{
    int i, j, k = 0;
    unsigned errors = 0, warnings = 0, notes = 0;
    unsigned long start1, start2, end1, end2, rs;
    unsigned long xgm_start, xgm_size, xgm_end;
    
    /* are all partitions mutually disjoint? (and disjoint with the bsl?) */
    /* do they start past zero and before end-of-disk? */
    for( i = 0; i < partitions; ++i ) {
	start1 = part_table[i].start;
	end1   = start1 + part_table[i].size;

	if (start1 < 1) {
	    printf( "Error: Partition %d starts before sector 1 (at %lu)\n",
		    i+1, start1 );
	    ++errors;
	}
	if (end1 > hd_size) {
	    printf( "Error: Partition %d ends after end of disk "
		    "(at %lu > %lu)\n", i+1, end1-1, hd_size-1 );
	    ++errors;
	}

	for( j = i+1; j < partitions; ++j ) {
	    start2 = part_table[j].start;
	    end2   = start2 + part_table[j].size;
	    if (end1 > start2 && end2 > start1) {
		printf( "Error: Partitions %d and %d overlap!\n", i+1, j+1 );
		++errors;
	    }
	}

	if (end1 > bsl_start && bsl_start+bsl_size > start1) {
	    printf( "Error: Partition %d overlaps with "
		    "the bad sector list!\n", i+1 );
	    ++errors;
	}
    }

    /* are all aux. rootsectors not contained in another partition or in the
     * bsl? And auxiliary root sectors must be before their data partitions */
    for( i = 0; i < partitions; ++i ) {
	rs = part_table[i].rootsec;
	if (!rs) continue;
	
	if (rs >= hd_size) {
	    printf( "Error: Partition %d has aux. root sector after end of "
		    "disk (at %lu > %lu)\n", i+1, rs, hd_size-1 );
	    ++errors;
	}

	if (rs >= part_table[i].start) {
	    printf( "Error: Partition %d's aux. root sector (%lu) is after "
		    "the partition start (%lu)\n",
		    i+1, rs, part_table[i].start );
	    ++errors;
	}
	
	for( j = 0; j < partitions; ++j ) {
	    if (i == j) continue;
	    if (inside_part( &part_table[j], rs )) {
		printf( "Error: Partition %d's aux. root sector is inside "
			"partition %d!\n", i+1, j+1 );
		++errors;
	    }
	}

	if (bsl_start <= rs && rs < bsl_start+bsl_size) {
	    printf( "Error: Partition %d's aux. root sector is inside "
		    "the bad sector list", i+1 );
	    ++errors;
	}
    }
    
    /* first primary slot may not be XGM */
    if (partitions > 0 && IS_EXTENDED(part_table[0])) {
	printf( "Error: First partition may not be extended "
		"(first primary slot must not be XGM)\n" );
	++errors;
    }
    
    /* only one XGM chain */
    for( i = j = 0; i < partitions; ++i ) {
	if (IS_EXTENDED(part_table[i])) {
	    if (i == 0 || !IS_EXTENDED(part_table[i-1]))
		++j;
	    k = i;
	}
    }
    if (j > 1) {
	printf( "Error: More than 1 (%d) XGM chains\n", j );
	++errors;
    }

    /* better no primary partitions after XGM chain */
    if (j > 0 && k != partitions-1) {
	printf( "Note: There are primary partitions after the XGM chain "
		"(unusual but ok).\n"
		" But maybe some weird TOS hd drivers can have problems "
		"with this.\n" );
	++notes;
    }

    if (j == 1) {
	/* does the XGM start past zero and before end-of-disk? */
	XGM_boundary( &xgm_start, &xgm_size );
	xgm_end = xgm_start + xgm_size;
	
	if (xgm_start < 1) {
	    printf( "Error: XGM partition starts before sector 1 (at %lu)\n",
		    xgm_start );
	    ++errors;
	}	
	if (xgm_end > hd_size) {
	    printf( "Error: XGM partition ends after end of disk "
		    "(at %lu > %lu)\n", xgm_end-1, hd_size-1 );
	    ++errors;
	}

	/* are the extended partitions ordered, and are primary partitions
	 * contained inside the XGM area? */
	end2 = 0; /* last end of extd */
	j = 1; /* are ordered */
	for( i = 0; i < partitions; ++i ) {
	    start1 = part_table[i].start;
	    end1   = start1 + part_table[i].size;
	    if (IS_EXTENDED(part_table[i])) {
		if (start1 < end2)
		    j = 0;
		else
		    end2 = end1;
	    }
	    else {
		if (xgm_end > start1 && end1 > xgm_start) {
		    printf( "Note: Primary partition %d inside XGM area!\n",
			    i+1 );
		    ++notes;
		}
	    }
	}
	if (!j) {
	    printf( "Note: Extended partitions are not ordered by disk "
		    "location\n" );
	    ++notes;
	}
    }
    
    /* check for invalid boot flags (more than one bit, flags set on extended
     * part.s */
    for( i = 0; i < partitions; ++i ) {
	j = part_table[i].flag;
	if (j == 0)
	    continue;
	if (!strcmp( atari_boot_type( j ), "???" )) {
	    printf( "Warning: Partition %d has invalid boot type (%02x)\n",
		    i+1, j );
	    ++warnings;
	}
	if (IS_EXTENDED(part_table[i])) {
	    printf( "Warning: Partition %d has boot type set but is extended\n"
		    " (booting from extended partitions is usually "
		    "not supported)\n", i+1 );
	    ++warnings;
	}
    }

    /* check if partition IDs look ok */
    for( i = j = 0; i < partitions; ++i ) {
	if (!is_valid_PID( part_table[i].id )) {
	    printf( "Warning: Partition %d has invalid ID (%s) "
		    "(not alphanumeric)\n", i+1, part_table[i].id );
	    ++warnings;
	}
	if (!(isupper(part_table[i].id[0]) &&
	      isupper(part_table[i].id[1]) &&
	      isupper(part_table[i].id[2]))) {
	    printf( "Note: Partition %d has unusual ID (%s) "
		    "(not three uppercase letters)\n", i+1, part_table[i].id );
	    ++notes;
	}
    }
    
    /* 'GEM' type partitions cannot be bigger than 32766 sectors */
    for( i = j = 0; i < partitions; ++i ) {
	if (strcmp( part_table[i].id, "GEM" ) == 0 &&
	    part_table[i].size > 32766) {
	    printf( "Warning: Partition %d has ID \"GEM\", but has more "
		    "than 32766 sectors.\n"
		    " Better use BGM.\n", i+1 );
	    ++warnings;
	}
    }
    
    /* warn if more than 15 part.s on SCSI/ACSI, 63 on IDE */
    if (partitions > 15 &&
	(!strncmp( disk_device, "/dev/sd", 7 ) ||
	 !strncmp( disk_device, "/dev/ad", 7 ))) {
	printf( "Note: Linux only supports 15 partitions on SCSI/ACSI devices "
		"(you have %d)\n", partitions );
	++notes;
    }
    if (partitions > 63 && (!strncmp( disk_device, "/dev/hd", 7 ))) {
	printf( "Note: Linux only supports 63 partitions on IDE devices "
		"(you have %d)\n", partitions );
	++notes;
    }

    printf( "\n" );
    printf( "%d errors (violations of partition format)\n", errors );
    printf( "%d warnings (things that may or may not work)\n", warnings );
    printf( "%d notes (basically ok, but at least unusual)\n", notes );
}

void write_table(void)
{
    int n_prim, n_ext, first_ext, last_ext;
    PARTITION XGM_part;

#ifndef ALPHA_VERSION
    /* run put_boot for alpha versions even if opened read-only; the actual
     * writing will not be avoided later */
    if (type_open == O_RDONLY) {
	printf( "%s is opened read-only; can't write partition table\n",
		disk_device );
	return;
    }
#endif
    
    table_infos( &n_prim, &n_ext, &first_ext, &last_ext );
    if (n_prim + (n_ext > 0) > 4) {
	printf( "Too many primary slots allocated. "
		"This should not happen!\n" );
	return;
    }

    if (n_ext > 0) {
	XGM_boundary( &XGM_part.start, &XGM_part.size );
	XGM_part.flag = XGM_flag;
	strcpy( XGM_part.id, "XGM" );
    }
    else
	first_ext = last_ext = -1;

    put_boot( first_ext, last_ext, &XGM_part );
}

/* Local Variables: */
/* tab-width: 8     */
/* End:             */
