/* partable.c: utilities for scanning the in-core partition table
 *
 * Copyright (C) 1995-97 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
 *               1996-97 Michael Schlueter <schlue00@marvin.informatik.uni-dortmund.de>
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation: either version 2 or
 * (at your option) any later version.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fdisk.h"
#include "partable.h"


/* sector inside partition? */
int inside_part( PARTITION *p, unsigned long sec )
{
    return( p->start <= sec && sec < p->start + p->size );
}

/* return number of primary partitions */
int n_primary( void )
{
    int i, n = 0;
    
    for( i = 0; i < partitions; ++i )
	if (!IS_EXTENDED(part_table[i]))
	    ++n;
    return( n );
}

/* in which primary slot the XGM will go? */
int XGM_slot( void )
{
    int i;
    
    for( i = 0; i < partitions && !IS_EXTENDED(part_table[i]); ++i )
	;
    return( i );
}

/* determine boundaries of XGM partition */
void XGM_boundary( unsigned long *start, unsigned long *size )
{
    unsigned long sta, end;
    int i;
    PARTITION *p;

    sta = hd_size;
    end = 0;
    for( i = 0; i < partitions; ++i ) {
	p = &part_table[i];
	if (IS_EXTENDED(*p)) {
	    if (p->start < sta)
		sta = p->start;
	    if (p->rootsec < sta)
		sta = p->rootsec;
	    if (p->start + p->size > end)
		end = p->start + p->size;
	}
    }

    *start = sta;
    *size  = end - sta;
}

void table_infos( int *n_prim, int *n_ext, int *first_ext, int *last_ext )
{
    int np = 0, ne = 0, fe = -1, le = -1;
    int i;
    
    for( i = 0; i < partitions; ++i ) {
	if (IS_EXTENDED(part_table[i])) {
	    ++ne;
	    if (fe < 0)
		fe = i;
	    le = i;
	}
	else
	    ++np;
    }

    if (n_prim) *n_prim = np;
    if (n_ext) *n_ext = ne;
    if (first_ext) *first_ext = fe;
    if (last_ext) *last_ext = le;
}

/* return canonical number for a partition (by disk location order) */
int canonical_number( unsigned long start )
{
    int i, new_num = 0;
    
    for( i = 0; i < partitions; ++i ) {
	if (part_table[i].start < start)
	    ++new_num;
    }
    return( new_num );
}

/* Can a new partition NEW_NUM be a primary partition given the rest of
 * layout? If not, return reason string in WHY if not NULL. */
int primary_possible( int new_num, int n_prim, int n_ext, int first_ext,
		      int last_ext, char **why )
{
    if (n_prim + (n_ext > 0) >= 4) {
	if (why)
	    *why = "no primary slots available";
	return( 0 );
    }
    if (n_ext > 0 &&
	new_num >= first_ext + (new_num <= first_ext) &&
	new_num <= last_ext  + (new_num <= last_ext )) {
	if (why)
	    *why = "would break XGM chain into two";
	return( 0 );
    }
    return( 1 );
}

/* Can a new partition NEW_NUM be an extended partition given the rest of
 * layout? If not, return reason string in WHY if not NULL. */
int extended_possible( int new_num, int n_prim, int n_ext, int first_ext,
		       int last_ext, char **why )
{
    if (new_num == 0) {
	if (why)
	    *why = "first partition may not be extended";
	return( 0 );
    }
    if (n_ext == 0 && n_prim == 4) {
	if (why)
	    *why = "no primary slot for XGM available";
	return( 0 );
    }
    if (n_ext > 0 &&
	(new_num < first_ext + (new_num <= first_ext) - 1 ||
	 new_num > last_ext  + (new_num <= last_ext ) + 1)) {
	if (why)
	    *why = "not in or next to existing XGM area";
	return( 0 );
    }
    return( 1 );
}

/* return number of unallocated sectors starting at 'sec' */
int free_sectors_at( unsigned long sec )
{
    int i;
    unsigned long n, start, end, rs;

    /* assume everything free to end of disk... */
    n = hd_size - sec;

    /* scan partitions and cut free area if necessary */
    for( i = 0; i < partitions; ++i ) {
	start = part_table[i].start;
	end   = start + part_table[i].size;
	if (start < sec+n && end > sec) {
	    if (start <= sec)
		/* overlapping 'sec' -> 'sec' not free at all */
		return( 0 );
	    else
		n = start - sec;
	}
	/* consider aux rs if part. is extended */
	rs = part_table[i].rootsec;
	if (rs) {
	    if (rs == sec)
		return( 0 );
	    else if (rs > sec && rs < sec+n)
		n = rs - sec;
	}
    }

    /* also compare with bad sector list */
    if (bsl_start < sec+n && bsl_start+bsl_size > sec) {
	if (bsl_start <= sec)
	    return( 0 );
	else
	    n = bsl_start - sec;
    }
    
    return( n );
}

/* return number of unallocated sectors before (and excluding) 'sec' */
int free_sectors_before( unsigned long sec )
{
    int i;
    unsigned long s, start, end,  rs;

    /* assume sectors [1..sec-1] free */
    s = 1;

    /* scan partitions and cut free area if necessary */
    for( i = 0; i < partitions; ++i ) {
	start = part_table[i].start;
	end   = start + part_table[i].size;
	if (start < sec && end > s) {
	    if (end >= sec)
		/* overlapping 'sec'-1 -> 'sec'-1 not free at all */
		return( 0 );
	    else
		s = end;
	}
	/* consider aux rs if part. is extended */
	rs = part_table[i].rootsec;
	if (rs) {
	    if (rs == sec-1)
		return( 0 );
	    else if (rs < sec && rs >= s)
		s = rs + 1;
	}
    }

    /* also compare with bad sector list */
    end = bsl_start + bsl_size;
    if (bsl_start < sec && end > s) {
	if (end >= sec)
	    return( 0 );
	else
	    s = end;
    }
    
    return( sec - s );
}

/* return number of allocated sectors starting at 'sec' */
int allocated_sectors_at( unsigned long sec )
{
    int i;
    unsigned long end;

    end = sec;
  rescan:
    for( i = 0; i < partitions; ++i ) {
	if (inside_part( &part_table[i], end )) {
	    /* current end is inside this partition -> new end of allocated
	     * region is end of this partition; need to rescan table from the
	     * beginning now, since part.s need not be in ascending order! */
	    end = part_table[i].start + part_table[i].size;
	    goto rescan;
	}
	if (end == part_table[i].rootsec) {
	    /* current end is aux rs for some extended partition */
	    ++end;
	    goto rescan;
	}
    }

    /* also bad sector list allocates sectors... */
    if (bsl_start <= end && end < bsl_start + bsl_size) {
	end = bsl_start + bsl_size;
	goto rescan;
    }
    
    return( end - sec );
}

/* Local Variables: */
/* tab-width: 8     */
/* End:             */
